# Copyright 2014, 2015, 2016, 2017, 2018, 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


package MyLineSegments;
use 5.010;
use strict;
use warnings;
use Carp 'croak';
use Clone::PP;
use List::Util 'min','max','sum0';

# uncomment this to run the ### lines
# use Smart::Comments;

sub new {
  my $class = shift;
  return bless { @_ }, $class;
}
sub copy {
  my ($self) = @_;
  return Clone::PP::clone($self);
}
sub bbox {
  my ($self) = @_;
  ### bbox(): $self
  return ($self->{'x_min'},
          $self->{'y_min'},
          $self->{'x_max'},
          $self->{'y_max'});
}
sub add {
  my ($self, $x1,$y1, $x2,$y2) = @_;
  unless ($x1 == $x2 || $y1 == $y2) {
    croak "Not vertical/horizontal line  $x1,$y1  $x2,$y2";
  }
  if (! defined $self->{'x_min'}) {
    $self->{'x_min'} = $self->{'x_max'} = $x1;
    $self->{'y_min'} = $self->{'y_max'} = $y1;
  }
  $self->{'x_min'} = min($self->{'x_min'}, $x1, $x2);
  $self->{'y_min'} = min($self->{'y_min'}, $y1, $y2);
  $self->{'x_max'} = max($self->{'x_max'}, $x1, $x2);
  $self->{'y_max'} = max($self->{'y_max'}, $y1, $y2);

  if ($x1 > $x2 || $y1 > $y2) {
    ($x1,$y1, $x2,$y2) = ($x2,$y2, $x1,$y1);
  }
  foreach my $x ($x1 .. $x2-1) {
    $self->{'horiz'}->{$x}{$y1} = 1;
  }
  foreach my $y ($y1 .. $y2-1) {
    $self->{'vert'}->{$x1}{$y} = 1;
  }
}

# $segs->add_cycle($x,$y, $x,$y, $x,$y, ...)
sub add_cycle {
  my $self = shift;
  return unless @_;
  my $prev_x = my $orig_x = shift;
  my $prev_y = my $orig_y = shift;
  while (@_) {
    my $x = shift;
    my $y = shift;
    $self->add($prev_x,$prev_y, $x,$y);
    ($prev_x,$prev_y) = ($x,$y);
  }
  $self->add($prev_x,$prev_y, $orig_x,$orig_y);
}

# FIXME: doesn't update bounding box
sub remove {
  my ($self, $x1,$y1, $x2,$y2) = @_;
  unless ($x1 == $x2 || $y1 == $y2) {
    croak "Not vertical/horizontal line";
  }

  if ($x1 > $x2 || $y1 > $y2) {
    ($x1,$y1, $x2,$y2) = ($x2,$y2, $x1,$y1);
  }
  foreach my $x ($x1 .. $x2-1) {
    delete $self->{'horiz'}->{$x}{$y1};
  }
  foreach my $y ($y1 .. $y2-1) {
    delete $self->{'vert'}->{$x1}{$y};
  }
}

sub xor {
  my ($self, $x1,$y1, $x2,$y2) = @_;
  unless ($x1 == $x2 || $y1 == $y2) {
    croak "Not vertical/horizontal line";
  }
  if ($x1 > $x2 || $y1 > $y2) {
    ($x1,$y1, $x2,$y2) = ($x2,$y2, $x1,$y1);
  }
  foreach my $x ($x1 .. $x2-1) {
    if ($self->is_traversed($x,$y1, $x+1,$y1)) {
      $self->remove($x,$y1, $x+1,$y1);
    } else {
      $self->add($x,$y1, $x+1,$y1);
    }
  }
  foreach my $y ($y1 .. $y2-1) {
    if ($self->is_traversed($x1,$y, $x1,$y+1)) {
      $self->remove($x1,$y, $x1,$y+1);
    } else {
      $self->add($x1,$y, $x1,$y+1);
    }
  }
}

sub _equal_with_undef {
  my ($a,$b) = @_;
  return (! defined $a && ! defined $b)
    || (defined $a && defined $b && $a == $b);
}

# sub equal {
#   my ($self, $other) = @_;
#
#   foreach my $field ('x_min', 'y_min', 'x_max', 'y_max') {
#     _equal_with_undef($self->{$field}, $other->{$field})
#       or return 0;
#   }
#
#   foreach my $x (keys %{$self->{'horiz'}}) {
#     my $column = $self->{'horiz'}->{$x};
#     foreach my $y (keys %$column) {
#     }
#   }
# }

# add the segments of another MyLineSegments
sub add_segments {
  my ($self, $segments) = @_;
  foreach my $line ($segments->lines) {
    my ($x1,$y1, $x2,$y2) = @$line;
    $self->add($x1,$y1, $x2,$y2);
  }
}
# remove the segments of another MyLineSegments
sub remove_segments {
  my ($self, $segments) = @_;
  foreach my $line ($segments->lines) {
    my ($x1,$y1, $x2,$y2) = @$line;
    $self->remove($x1,$y1, $x2,$y2);
  }
}

sub translate {
  my ($self, $dx, $dy) = @_;
  ### translate(): "$dx,$dy on ".$self->segments_string
  foreach my $href ($self->{'horiz'}, $self->{'vert'}) {
    my @new;
    while (my ($x,$subhref) = each %$href) {
      push @new, $x+$dx, $subhref;
    }
    %$href = @new;
  }
  foreach my $href ($self->{'horiz'}, $self->{'vert'}) {
    foreach my $subhref (values %$href) {
      my @new;
      foreach my $y (keys %$subhref) {
        push @new, $y+$dy, 1;
      }
      %$subhref = @new;
    }
  }
  $self->{'x_min'} += $dx; $self->{'y_min'} += $dy;
  $self->{'x_max'} += $dx; $self->{'y_max'} += $dy;

  ### now: $self
}

sub is_traversed {
  my ($self, $x1,$y1, $x2,$y2) = @_;
  unless ($x1 == $x2 || $y1 == $y2) {
    croak "Not vertical/horizontal line";
  }
  if ($x1 > $x2 || $y1 > $y2) {
    ($x1,$y1, $x2,$y2) = ($x2,$y2, $x1,$y1);
  }
  my $href = $self->{$x1 == $x2 ? 'vert' : 'horiz'};
  return exists $href->{$x1}{$y1};
}
sub is_visited {
  my ($self, $x,$y) = @_;
  ### is_visited() ...
  return (exists $self->{'vert'}->{$x}{$y}       # start of a vertical
          || exists $self->{'vert'}->{$x}{$y-1}  # end of a vertical
          || exists $self->{'horiz'}->{$x}{$y}   # start of a horizontal
          || exists $self->{'horiz'}->{$x-1}{$y} # end of a horizontal
         );
}
# Return a list of arrayrefs [$x1,$y1, $x2,$y2] which are line segments.
sub lines {
  my ($self) = @_;
  my ($x_min,$y_min, $x_max,$y_max) = $self->bbox;
  my @ret;

  if (defined $x_min) {
    foreach my $y ($y_min .. $y_max) {
      my ($x_from,$x_to);
      foreach my $x ($x_min .. $x_max) {
        if ($self->is_traversed($x,$y, $x+1,$y)) {
          if (! defined $x_from) {
            $x_from = $x;
          }
          $x_to = $x+1;
        } else {
          if (defined $x_from) {
            push @ret, [ $x_from,$y, $x_to,$y ];
            undef $x_from;
          }
        }
      }
    }

    foreach my $x ($x_min .. $x_max) {
      my ($y_from,$y_to);
      foreach my $y ($y_min .. $y_max) {
        if ($self->is_traversed($x,$y, $x,$y+1)) {
          if (! defined $y_from) {
            $y_from = $y;
          }
          $y_to = $y+1;
        } else {
          if (defined $y_from) {
            push @ret, [ $x,$y_from, $x,$y_to ];
            undef $y_from;
          }
        }
      }
    }
  }
  return @ret;
}

sub num_lines_hvt {
  my ($self) = @_;
  my @lines = $self->lines;
  my $verticals   = grep {$_->[0] == $_->[2]} @lines;  # x1==x2
  my $horizontals = grep {$_->[1] == $_->[3]} @lines;  # y1==y2
  return ($horizontals, $verticals, scalar(@lines));
}

sub _xy_string_cmp {
  my ($a,$b) = @_;
  my ($ax,$ay) = split /,/,$a;
  my ($bx,$by) = split /,/,$b;
  return ($ax <=> $bx || $ay <=> $ay);
}

# use Smart::Comments;

# Return a list of arrayrefs [$x1,$y1, $x2,$y2, ..., $xn,$yn] which are line
# runs
sub lines_joined {
  my ($self) = @_;
  my @lines = $self->lines;
  # print "separate lines ",scalar(@lines),"\n";

  my %start_hash;
  my %end_hash;
  my %all;

  my $add = sub {
    my ($line) = @_;
    ### add: $line
    {
      my $xy = "$line->[0],$line->[1]";
      push @{$start_hash{$xy}}, $line;
      $all{$xy} = 1;
    }
    {
      my $xy = "$line->[-2],$line->[-1]";
      push @{$end_hash{$xy}}, $line;
      $all{$xy} = 1;
    }
  };

  my $remove_from_aref = sub {
    my ($aref, $elem) = @_;
    my $length = @$aref;
    @$aref = grep { $_ != $elem } @$aref;
    if (@$aref != $length-1) {
      die "oops, bad remove";
    }
  };

  my $remove = sub {
    my ($line) = @_;
    ### remove: $line
    {
      my $xy = "$line->[0],$line->[1]";
      my $start_list = $start_hash{$xy};
      $remove_from_aref->($start_list, $line);
      ### remove to start_list: $start_list
    }
    {
      my $xy = "$line->[-2],$line->[-1]";
      my $end_list = $end_hash{$xy};
      $remove_from_aref->($end_list, $line);
      ### remove to start_list: $end_list
    }
  };

  foreach my $line (@lines) {
    $add->($line);
  }
  @lines = ();
  ### %start_hash
  ### %end_hash

  my @keys = keys %all;
  @keys = sort \&_xy_string_cmp, @keys;
  foreach my $xy (@keys) {
    ### $xy
    my $start_list = $start_hash{$xy} // [];
    my $end_list   = $end_hash{$xy}   // [];
    ### $start_list
    ### $end_list

    for (;;) {
      my $from_rev = 1;
      my $from_line = $start_list->[0]
        // do { $from_rev = 0; $end_list->[0] }
        // last;
      ### $from_rev
      ### $from_line
      $remove->($from_line);

      my $to_rev = 0;
      my $to_line = $start_list->[0]
        // do { $to_rev = 1; $end_list->[0] };
      if (! $to_line) {
        $add->($from_line);
        last;
      }
      ### $to_rev
      ### $to_line

      $remove->($to_line);

      if ($from_rev) {
        $from_line = _reverse_pairs($from_line);
      }
      if ($to_rev) {
        $to_line = _reverse_pairs($to_line);
      }

      ### join ...
      splice @$from_line, -2;
      my $new_line = [ @$from_line, @$to_line ];
      $add->($new_line);
    }
  }

  ### joins done ...

  foreach my $xy (@keys) {
    if (my $start_list = $start_hash{$xy}) {
      foreach my $line (@$start_list) {
        push @lines, $line;
        $remove->($line);
      }
    }
    if (my $end_list = $end_hash{$xy}) {
      foreach my $line (@$end_list) {
        push @lines, $line;
        $remove->($line);
      }
    }
  }

  # sort ends to go east and north
  @lines = map {
    my $line = $_;
    (($line->[0] <=> $line->[-2] || $line->[1] <=> $line->[-1]) <= 0)
      ? $line
      : _reverse_pairs($line)
    } @lines;

  $self->{'lines_joined_count_points'}
    = sum0(map {_lines_aref_count_points($_)} @lines);

  #  print "joined lines ",scalar(@lines),"\n";
  return @lines;
}
sub _lines_aref_count_points {
  my ($line) = @_;
  return scalar(@$line)/2 - ($line->[0] == $line->[-2]
                             && $line->[1] == $line->[-1]);
}
sub lines_joined_count_points {
  my ($self) = @_;
  if (! defined $self->{'lines_joined_count_points'}) {
    $self->lines_joined;
  }
  return $self->{'lines_joined_count_points'};
}


sub _reverse_pairs {
  my ($aref) = @_;
  my @ret = reverse @$aref;
  for (my $i = 0; $i <= $#ret; $i += 2) {
    ($ret[$i], $ret[$i+1]) = ($ret[$i+1], $ret[$i]);
  }
  return \@ret;
}

# Return a list of arrayrefs [$x1,$y1, $x2,$y2] which are individual unit
# segments.
sub segments {
  my ($self) = @_;
  my ($x_min,$y_min, $x_max,$y_max) = $self->bbox;
  my @ret;

  if (defined $x_min) {
    foreach my $y ($y_min .. $y_max) {
      foreach my $x ($x_min .. $x_max) {
        if ($self->is_traversed($x,$y, $x+1,$y)) {
          push @ret, [ $x,$y, $x+1,$y ];
        }
      }
    }
    foreach my $x ($x_min .. $x_max) {
      my ($y_from,$y_to);
      foreach my $y ($y_min .. $y_max) {
        if ($self->is_traversed($x,$y, $x,$y+1)) {
          push @ret, [ $x,$y, $x,$y+1 ];
        }
      }
    }
  }
  return @ret;
}

# $seg is an arrayref [$x1,$y1, $x2,$y2] like from $self->segments
sub is_surrounded_R {
  my ($self, $seg) = @_;
  my ($x,$y, $x2,$y2) = @$seg;
  my $dx = $x2 - $x;
  my $dy = $y2 - $y;
  my $rx = $dy;
  my $ry = -$dx;
  #
  # x+dx,y+dy ----> x+dx+rx, y+dy+ry
  #     |
  #     |
  #    x,y  -------> x+rx,y+ry
  #         rx,ry
  #
  return ($self->is_traversed($x,$y,         $x+$rx,$y+$ry)
          &&
          $self->is_traversed($x+$dx,$y+$dy, $x+$rx+$dx,$y+$ry+$dy)
          &&
          $self->is_traversed($x+$rx,$y+$ry, $x+$rx+$dx,$y+$ry+$dy));
}
sub is_surrounded_L {
  my ($self, $seg) = @_;
  my ($x,$y, $x2,$y2) = @$seg;
  my $dx = $x2 - $x;
  my $dy = $y2 - $y;
  my $lx = -$dy;
  my $ly = $dx;
  #
  # x+dx+lx, y+dy+ly <--- x+dx,y+dy
  #                           |
  #                           |
  #    x+lx,y+ly  <--------- x,y
  #                  lx,ly
  #
  return ($self->is_traversed($x,$y,         $x+$lx,$y+$ly)
          &&
          $self->is_traversed($x+$dx,$y+$dy, $x+$lx+$dx,$y+$ly+$dy)
          &&
          $self->is_traversed($x+$lx,$y+$ly, $x+$lx+$dx,$y+$ly+$dy));
}

# Return a list of coordinates $x1,$y1, $x2,$y2, ... which are the segments
# as a single cycle.  If a single cycle cannot be formed then croak.
sub cycle_coordinates {
  my ($self) = @_;
  my @lines = $self->lines;
  if (! @lines) { return (); }

  my @ret = @{shift @lines};
 LINE: while (@lines) {
    my $x = $ret[-2];
    my $y = $ret[-1];

    ### ret: join(',',@ret)
    ### seek: "xy=$x,$y"

    foreach my $i (0 .. $#lines) {
      if ($x == $lines[$i]->[0] && $y == $lines[$i]->[1]) {
        push @ret, $lines[$i]->[2], $lines[$i]->[3];
        ### remove: "$i  ".join(',',@{$lines[$i]})
        splice @lines, $i, 1;
        next LINE;
      }
      if ($x == $lines[$i]->[2] && $y == $lines[$i]->[3]) {
        push @ret, $lines[$i]->[0], $lines[$i]->[1];
        ### remove: "$i  ".join(',',@{$lines[$i]})
        splice @lines, $i, 1;
        next LINE;
      }
    }

    # print "remaining lines:\n";
    # foreach my $line (@lines) {
    #   my ($x1,$y1, $x2,$y2) = @$line;
    #   print "  ($x1,$y1)--($x2,$y2)\n";
    # }
    # print "\n";

    croak "Not a single cycle (dead end at $x,$y)";
  }
  return @ret;
}

sub segments_string {
  my ($self) = @_;
  my @lines = $self->segments;
  my @line_strings = map { my ($x1,$y1, $x2,$y2) = @$_;
                           "[$x1,$y1 to $x2,$y2]" } @lines;
  return join(', ', @line_strings);
}

# return list ($x,$y, $x,$y, ...)
sub points {
  my ($self) = @_;
  ### points() ...
  my ($x_min,$y_min, $x_max,$y_max) = $self->bbox;
  my @ret;

  if (defined $x_min) {
    foreach my $y ($y_min .. $y_max) {
      foreach my $x ($x_min .. $x_max) {
        if ($self->is_traversed($x,$y, $x+1,$y)
            || $self->is_traversed($x,$y, $x-1,$y)) {
          push @ret, [ $x,$y ];
        }
      }
    }
    foreach my $x ($x_min .. $x_max) {
      my ($y_from,$y_to);
      foreach my $y ($y_min .. $y_max) {
        if ($self->is_traversed($x,$y, $x,$y+1)
            || $self->is_traversed($x,$y, $x,$y-1)) {
          push @ret, [ $x,$y ];
        }
      }
    }
  }
  ### $self
  ### @ret
  return @ret;
}

sub print_ascii {
  my ($self) = @_;
  my ($x_min,$y_min, $x_max,$y_max) = $self->bbox;
  my $str = '';
  foreach my $y (reverse $y_min-1 .. $y_max+1) {
    foreach my $x ($x_min .. $x_max+1) {
      $str .= $self->is_traversed($x,$y, $x,$y+1) ? "|" : " ";
      $str .= "   ";
    }
    $str .= "\n";
    foreach my $x ($x_min .. $x_max+1) {
      $str .= $self->is_visited($x,$y) ? "*" : " ";
      $str .= $self->is_traversed($x,$y, $x+1,$y) ? "---" : "   ";
    }
    $str .= "\n";
  }
  $str =~ s/ +\n/\n/g;
  $str =~ s/^\n+//;
  $str =~ s/\n+$/\n/;
  # $str =~ s/\n/.\n/;
  print $str;
}

sub length {
  my ($self) = @_;
  my $length = 0;
  foreach my $line ($self->lines) {
    my ($x1,$y1, $x2,$y2) = @$line;
    $length += ($x2 - $x1) + ($y2 - $y1);
  }
  return $length;
}

sub is_subset_of {
  my ($self, $superset_segments) = @_;
  my $diff = $self->copy;
  $diff->remove_segments($superset_segments);
  return $diff->length == 0;
}

sub is_disjoint {
  my ($self, $other_segments) = @_;
  my $diff = $self->copy;
  $diff->remove_segments($other_segments);
  $diff->length == $self->length or return 0;

  $diff = $other_segments->copy;
  $diff->remove_segments($self);
  return $diff->length == $other_segments->length;
}


#------------------------------------------------------------------------------
# printwrap bits

sub _transform_noop {
  return @_;
}
sub _stringize_xy {
  return undef;
}

use Smart::Comments;
sub printwrap {
  my ($segments, %options) = @_;
  my $transform = delete $options{'transform'} || \&_transform_noop;
  my $stringize = delete $options{'stringize'} || \&_stringize_xy;
  my $use_increments = (defined delete $options{'use_increments'}
                        ? $options{'use_increments'} : 1);
  croak "unrecognised options" if %options;

  require MyPrintwrap;
  my $separator = '';
  foreach my $line ($segments->lines_joined) {
    my @points;
    my ($prev_x,$prev_y);
    for (my $i = 0; $i < $#$line; $i+=2) {
      my $x = $line->[$i];
      my $y = $line->[$i+1];
      if ($i == $#$line-1 && $x == $line->[0] && $y == $line->[1]) {
        push @points, 'cycle';
        next;
      }
      ($x,$y) = $transform->($x,$y);
      my $str = $stringize->($x,$y);
      my $cstr = $str // "($x,$y)";
      if (defined $prev_x && ! defined $str) {
        my $dx = $x - $prev_x;
        my $dy = $y - $prev_y;
        my $dstr = "+($dx,$dy)";
        if (CORE::length($dstr) < CORE::length($cstr)) {
          push @points, $dstr;
          next; # $prev_x,$prev_y unmoved
        }
      }
      if (defined $str) {
        undef $prev_x;
        undef $prev_y;
      } elsif ($use_increments) {
        $prev_x = $x;
        $prev_y = $y;
      }
      push @points, $cstr;
    }

    my $join = $separator;
    foreach my $p (@points) {
      MyPrintwrap::printwrap("$join$p");
      $join = '--';
    }
    $separator = ' ';
  }
}

1;
__END__
