#!/usr/bin/perl -w

# Copyright 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use Tie::IxHash;
use Math::PlanePath::R5DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyR5Dragon 'xy_rotate_plus90','rect_corners_sort','chamfer',;
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 3;

print "  % generated by tools/$FindBin::Script\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::R5DragonCurve->new;

my ($n_lo, $n_hi) = $path->level_to_n_range($level);
my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
### level: "$level  n_hi=$n_hi"

my $n_start = 0*$n_hi;
my $n_end   = 2*$n_hi;


# JN(k) = 1/4 * (3*5^k + 1)
sub JN {
  my ($k) = @_;
  return (3*5**$k + 1) /4;
}
# JNother(k) = 1/4 * (7*5^k - 3)
sub JNother {
  my ($k) = @_;
  return (7*5**$k - 3) /4;
}

sub mul_iplus1 {
  my ($x,$y) = @_;
  # (x,y) + rot+90(x,y), is multiply (i+1)
  return ($x - $y,
          $y + $x);
}
sub div_iplus1 {
  my ($x,$y) = @_;
  # (x,y) + rot+90(x,y), is multiply (1-i)
  return (($x + $y)/2,
          ($y - $x)/2);
}

{
  my ($x,$y) = $path->n_to_xy($n_start);
  print "  \\coordinate (start) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($n_end);
  print "  \\coordinate (end) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($n_start + $n_hi);
  print "  \\node (J) at ($x,$y) [my small dot];\n";
}
{
  my ($x,$y) = $path->n_to_xy(JN($level));
  print "  \\coordinate (J-end) at ($x,$y);\n";
}

sub xyxy_is_touch {
  my ($x,$y) = @_;

  my @n_list = $path->xy_to_n_list($x,$y);
  @n_list = grep {$n_start <= $_ && $_ < $n_end} @n_list;
  return @n_list > 1;
}

sub n_is_left_enclosed {
  my ($n) = @_;
  my ($x,$y) = $path->n_to_xy($n);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  my $x2 = $x + $dx;
  my $y2 = $y + $dy;
  foreach (1 .. 3) {
    ($x,$y) = ($x2,$y2);
    ($dx,$dy) = xy_rotate_plus90($dx,$dy);
    $x2 += $dx;
    $y2 += $dy;
    if (! defined $path->xyxy_to_n_either($x,$y, $x2,$y2)) {
      return 0;
    }
  }
  return 1;
}

my $tree_segments = MyLineSegments->new;
tie my %nodes, 'Tie::IxHash';

$Printwrap = 2;
printwrap("\\draw[my r5dragon] ");
my $join = '';
$Printwrap += 2;
foreach my $n ($n_start .. $n_end) {
  my ($x,$y) = $path->n_to_xy($n);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  my $is_touch = ($n==$n_start || xyxy_is_touch($x,$y));

  my ($prev_dx,$prev_dy) = $path->n_to_dxdy($n-1);
  if (defined $prev_dx && $is_touch) {
    my $sx = $x . chamfer(-$prev_dx);
    my $sy = $y . chamfer(-$prev_dy);
    my $tx = $x . chamfer($dx);
    my $ty = $y . chamfer($dy);
    printwrap("$join($sx,$sy)");
    printwrap("--($tx,$ty)");
  } else {
    printwrap("$join($x,$y)");
  }
  $join = '--';
}
printwrap(";\n");

my ($prev_cx,$prev_cy);
foreach my $n (JN($level) .. JNother($level)-1) {
  my ($x,$y) = $path->n_to_xy($n);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  if (n_is_left_enclosed($n)) {
    my ($ldx,$ldy) = xy_rotate_plus90($dx,$dy);
    my ($cx,$cy) = rect_corners_sort($x,$y, $x+$dx+$ldx, $y+$dy+$ldy);
    $nodes{"$cx,$cy"}++;
    ($cx,$cy) = mul_iplus1($cx,$cy);
    if (defined $prev_cx) {
      $tree_segments->add($cx,$cy, $prev_cx,$prev_cy);
    }
    ($prev_cx,$prev_cy) = ($cx,$cy);
  } else {
    last;
  }
}

$Printwrap = 2;
printwrap("\\draw[my tree] ");
$Printwrap += 2;
$tree_segments->printwrap (transform => sub {
                             my ($x,$y) = @_;
                             ($x,$y) = div_iplus1($x,$y);
                             return ($x+.5, $y+.5);
                           });
printwrap(";\n");

$Printwrap = 2;
printwrap("\\path foreach \\p in {");
$Printwrap = 4;
foreach my $cxcy (keys %nodes) {
  my ($cx,$cy) = split /,/, $cxcy;
  printwrap("($cx+.5,$cy+.5),");
}
printwrap("} ");
printwrap("{ \\p node[my dot] };\n");

my $num_nodes = keys %nodes;
print "% GP-Test-Last  JAV($level) == $num_nodes   /* vertices */\n";
print "  % end generated\n";
exit 0;
