#!/usr/bin/perl -w

# Copyright 2016, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use List::Util 'sum';
use Math::PlanePath::R5DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyR5Dragon ':all';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 4;
if (@ARGV) { $level = shift @ARGV; }

print "  % generated by tools/$FindBin::Script $level\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::R5DragonCurve->new;
my ($n_lo, $n_hi) = $path->level_to_n_range($level);

my $n_start = 4*$n_hi;
my $n_end   = 8*$n_hi;
my ($x_start,$y_start) = $path->n_to_xy($n_start);

sub xyxy_is_touch {
  my ($x,$y) = @_;
  my @n_list = $path->xy_to_n_list($x,$y);
  @n_list = grep {$n_start <= $_ && $_ < $n_end} @n_list;
  return @n_list > 1;
}

my $tree_segments = MyLineSegments->new;
my %nodes;

$Printwrap = 2;
printwrap("\\draw[my r5quad] ");
my $join = '';
$Printwrap += 2;
foreach my $n ($n_start .. $n_end-1) {
  my ($x,$y) = $path->n_to_xy($n);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  my $is_touch = xyxy_is_touch($x,$y);

  $x -= $x_start;
  $y -= $y_start;

  if ($is_touch) {
    my ($prev_dx,$prev_dy) = $path->n_to_dxdy($n-1);
    my $sx = $x . chamfer(-$prev_dx);
    my $sy = $y . chamfer(-$prev_dy);
    my $tx = $x . chamfer($dx);
    my $ty = $y . chamfer($dy);
    printwrap("$join($sx,$sy)");
    printwrap("--($tx,$ty)");
  } else {
    printwrap("$join($x,$y)");
  }
  $join = '--';
}
printwrap("--cycle;\n");

foreach my $n ($n_start .. $n_end-1) {
  my ($x,$y) = $path->n_to_xy($n);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  my ($sx,$sy) = xy_rotate_plus90($dx,$dy);

  my $xopp = $x + $dx + $sx;
  my $yopp = $y + $dy + $sy;
  my ($x1,$y1, $x2,$y2) = rect_corners_sort ($x,$y, $xopp,$yopp);
  my $type = MyR5Dragon::square_consecutive_type_xy($x1,$y1, $n_start,$n_end);

  if ($type eq '0011L') {
    ### corner: "n=$n  $x1,$y1 to $x2,$y2  (from $x,$y $dx,$dy $sx,$sy)  $type"
    ### corner: "offset ".($x1-$x_start).",".($y1-$y_start)
    $nodes{"$x1,$y1"} = [$x1,$y1];
  }
}

$Printwrap = 2;
printwrap("\\path foreach \\p in {");
$Printwrap += 2;
foreach my $p (sort values %nodes) {
  my ($x,$y) = @$p;
  $x -= $x_start;
  $y -= $y_start;
  ### node corner: "$x,$y"
  $x += .5;
  $y += .5;
  printwrap("($x,$y),");
}
printwrap("}");
printwrap(" { \\p node[my dot] };\n");


# GP-Test  (x+I*y)*(1+I) == (x-y) + I*(x+y)
sub mul_iplus1 {
  my ($x,$y) = @_;
  return ($x - $y,
          $y + $x);
}

# GP-Test  (x+I*y)/(1+I) == (x+y)/2 + I*(y-x)/2
sub div_iplus1 {
  my ($x,$y) = @_;
  return (($x + $y)/2,
          ($y - $x)/2);
}

my @diag4_to_dx = (1,-1,-1,1);
my @diag4_to_dy = (1,1,-1,-1);

foreach my $p (values %nodes) {
  my ($x,$y) = @$p;
  foreach my $diag4 (0 .. 3) {
    my $dx = $diag4_to_dx[$diag4];
    my $dy = $diag4_to_dy[$diag4];
    my $x2 = $x + $dx;
    my $y2 = $y + $dy;
    if (exists $nodes{"$x2,$y2"}) {
      my ($x,$y)   = mul_iplus1($x,$y);
      my ($x2,$y2) = mul_iplus1($x2,$y2);
      $tree_segments->add($x,$y, $x2,$y2);
    }
  }
}

$Printwrap = 2;
printwrap("\\draw[my line] ");
$Printwrap += 2;
$tree_segments->printwrap(transform => sub {
                            my ($x,$y) = @_;
                            ($x,$y) = div_iplus1($x,$y);
                            $x -= $x_start;
                            $y -= $y_start;
                            return ($x+.5, $y+.5);
                          });
printwrap(";");
my $count = $tree_segments->lines_joined_count_points;
printwrap(" % $count drawing points\n");

my $num_nodes = scalar(keys %nodes);
print "  % $num_nodes vertices\n";

print "  % end generated\n";
exit 0;
